tables_sql = [
"""CREATE TABLE care_plans (
  id VARCHAR(45) PRIMARY KEY,
  request_method VARCHAR(7),
  resource_type VARCHAR(50),
  resource_id VARCHAR(36),
  status VARCHAR(25),
  intent VARCHAR(25),
  category JSONB,
  subject_reference VARCHAR(45),
  encounter_reference VARCHAR(45),
  period_start TIMESTAMP WITH TIME ZONE,
  period_end TEXT,
  care_team JSONB,
  addresses JSONB,
  activity JSONB
);""",
"""CREATE TABLE claims (
  id VARCHAR(45) PRIMARY KEY,
  request_method VARCHAR(7),
  resource_type VARCHAR(50),
  resource_id VARCHAR(36),
  status VARCHAR(25),
  type_coding JSONB,
  use VARCHAR(25),
  patient_reference VARCHAR(45),
  patient TEXT,
  billable_period_start TIMESTAMP WITH TIME ZONE,
  billable_period_end TIMESTAMP WITH TIME ZONE,
  created TIMESTAMP WITH TIME ZONE,
  provider_reference TEXT,
  provider TEXT,
  priority_coding JSONB,
  facility_reference TEXT,
  facility TEXT,
  diagnosis JSONB,
  supporting_info JSONB,
  procedure JSONB,
  prescription_reference VARCHAR(45),
  insurance JSONB,
  item JSONB,
  total_value TEXT,
  total_currency VARCHAR(4)
);""",
"""CREATE TABLE conditions (
  id VARCHAR(45) PRIMARY KEY,
  request_method VARCHAR(7),
  resource_type VARCHAR(50),
  resource_id VARCHAR(36),
  clinical_status JSONB,
  verification_status JSONB,
  category JSONB,
  code_coding JSONB,
  code TEXT,
  subject_reference VARCHAR(45),
  encounter_reference VARCHAR(45),
  onset_date_time TIMESTAMP WITH TIME ZONE,
  recorded_date TIMESTAMP WITH TIME ZONE
);""",
"""CREATE TABLE diagnostic_reports (
  id VARCHAR(45) PRIMARY KEY,
  request_method VARCHAR(7),
  resource_type VARCHAR(50),
  resource_id VARCHAR(36),
  status VARCHAR(25),
  category JSONB,
  code_coding JSONB,
  code TEXT,
  subject_reference VARCHAR(45),
  encounter_reference VARCHAR(45),
  effective_date_time TEXT,
  issued_date_time TEXT,
  performer JSONB,
  result JSONB,
  presented_form JSONB
);""",
"""CREATE TABLE document_references (
  id VARCHAR(45) PRIMARY KEY,
  request_method VARCHAR(7),
  resource_type VARCHAR(50),
  resource_id VARCHAR(36),
  status VARCHAR(25),
  type_coding JSONB,
  category JSONB,
  subject_reference VARCHAR(45),
  date TIMESTAMP WITH TIME ZONE,
  author JSONB,
  custodian_reference TEXT,
  custodian TEXT,
  content JSONB,
  context_encounter JSONB,
  context_period_start TIMESTAMP WITH TIME ZONE,
  context_period_end TIMESTAMP WITH TIME ZONE
);""",
"""CREATE TABLE encounters (
  id VARCHAR(45) PRIMARY KEY,
  request_method VARCHAR(7),
  resource_type VARCHAR(50),
  resource_id VARCHAR(36),
  status VARCHAR(25),
  class_code VARCHAR(10),
  type JSONB,
  subject_reference VARCHAR(45),
  subject TEXT,
  participant JSONB,
  period_start TIMESTAMP WITH TIME ZONE,
  period_end TIMESTAMP WITH TIME ZONE,
  reason_code JSONB,
  hospitalization_discharge_disposition_coding JSONB,
  hospitalization_discharge_disposition TEXT,
  location JSONB,
  service_provider_reference TEXT,
  service_provider TEXT
);""",
"""CREATE TABLE explanation_of_benefits (
  id VARCHAR(45) PRIMARY KEY,
  request_method VARCHAR(7),
  resource_type VARCHAR(50),
  resource_id VARCHAR(36),
  contained JSONB,
  identifier JSONB,
  status VARCHAR(25),
  type_coding JSONB,
  use VARCHAR(25),
  patient_reference VARCHAR(45),
  billable_period_start TIMESTAMP WITH TIME ZONE,
  billable_period_end TIMESTAMP WITH TIME ZONE,
  created TIMESTAMP WITH TIME ZONE,
  insurer TEXT,
  provider_reference TEXT,
  referral_reference TEXT,
  facility_reference TEXT,
  facility TEXT,
  claim_reference VARCHAR(45),
  outcome TEXT,
  care_team JSONB,
  diagnosis JSONB,
  insurance JSONB,
  item JSONB,
  total JSONB,
  payment_amount_value TEXT,
  payment_amount_currency VARCHAR(4)
);""",
"""CREATE TABLE medication_requests (
  id VARCHAR(45) PRIMARY KEY,
  request_method VARCHAR(7),
  resource_type VARCHAR(50),
  resource_id VARCHAR(36),
  status VARCHAR(25),
  intent VARCHAR(25),
  medication_reference VARCHAR(45),
  medication_coding JSONB,
  medication TEXT,
  subject_reference VARCHAR(45),
  encounter_reference VARCHAR(45),
  authored_on TIMESTAMP WITH TIME ZONE,
  requester_reference TEXT,
  requester TEXT,
  reason_reference JSONB,
  dosage_instruction JSONB
);""",
"""CREATE TABLE patients (
  id VARCHAR(45) PRIMARY KEY,
  request_method VARCHAR(7),
  resource_type VARCHAR(50),
  resource_id VARCHAR(36),
  extension JSONB,
  identifier JSONB,
  name JSONB,
  contact JSONB,
  gender VARCHAR(10),
  birth_date VARCHAR(10),
  deceased_date_time TEXT,
  address JSONB,
  marital_status_coding JSONB,
  marital_status VARCHAR(20),
  multiple_birth VARCHAR(5),
  multiple_birth_amount VARCHAR(5),
  communication JSONB
);""",
"""CREATE TABLE procedures (
  id VARCHAR(45) PRIMARY KEY,
  request_method VARCHAR(7),
  resource_type VARCHAR(50),
  resource_id VARCHAR(36),
  status VARCHAR(25),
  code_coding JSONB,
  code TEXT,
  subject_reference VARCHAR(45),
  encounter_reference VARCHAR(45),
  performed_period_start TIMESTAMP WITH TIME ZONE,
  performed_period_end TIMESTAMP WITH TIME ZONE,
  location_reference TEXT,
  location TEXT,
  reason_reference JSONB
);"""
]


SCHEMAS = {
        'CarePlan': {
            'table_meta': {
                'table_name': 'care_plans',
            },
            'json_schema': {
                'id': ['fullUrl'],
                'request_method': ['request', 'method'],
                'resource_type': ['resource', 'resourceType'],
                'resource_id': ['resource', 'id'],
                'status': ['resource', 'status'],
                'intent': ['resource', 'intent'],
                'category': ['resource', 'category'],
                'subject_reference': ['resource', 'subject', 'reference'],
                'encounter_reference': ['resource', 'encounter', 'reference'],
                'period_start': ['resource', 'period', 'start'],
                'period_end': ['resource', 'period', 'end'],
                'care_team': ['resource', 'careTeam'],
                'addresses': ['resource', 'addresses'],
                'activity': ['resource', 'activity']
            }
        },
        'Claim': {
            'table_meta': {
                'table_name': 'claims',
            },
            'json_schema': {
                'id': ['fullUrl'],
                'request_method': ['request', 'method'],
                'resource_type': ['resource', 'resourceType'],
                'resource_id': ['resource', 'id'],
                'status': ['resource', 'status'],
                'type_coding': ['resource', 'type', 'coding'],
                'use': ['resource', 'use'],
                'patient_reference': ['resource', 'patient', 'reference'],
                'patient': ['resource', 'patient', 'display'],
                'billable_period_start': ['resource', 'billablePeriod', 'start'],
                'billable_period_end': ['resource', 'billablePeriod', 'end'],
                'created': ['resource', 'created'],
                'provider_reference': ['resource', 'provider', 'reference'],
                'provider': ['resource', 'provider', 'display'],
                'priority_coding': ['resource', 'priority', 'coding'],
                'facility_reference': ['resource', 'facility', 'reference'],
                'facility': ['resource', 'facility', 'display'],
                'diagnosis': ['resource', 'diagnosis'],
                'supporting_info': ['resource', 'supportingInfo'],
                'procedure': ['resource', 'procedure'],
                'prescription_reference': ['resource', 'prescription', 'reference'],
                'insurance': ['resource', 'insurance'],
                'item': ['resource', 'item'],
                'total_value': ['resource', 'total', 'value'],
                'total_currency': ['resource', 'total' 'currency']
            }
        },
        'Condition': {
            'table_meta': {
                'table_name': 'conditions',
            },
            'json_schema': {
                'id': ['fullUrl'],
                'request_method': ['request', 'method'],
                'resource_type': ['resource', 'resourceType'],
                'resource_id': ['resource', 'id'],
                'clinical_status': ['resource', 'clinicalStatus', 'coding'],
                'verification_status': ['resource', 'verificationStatus', 'coding'],
                'category': ['resource', 'category'],
                'code_coding': ['resource', 'code', 'coding'],
                'code': ['resource', 'code', 'text'],
                'subject_reference': ['resource', 'subject', 'reference'],
                'encounter_reference': ['resource', 'encounter', 'reference'],
                'onset_date_time': ['resource', 'onsetDateTime'],
                'recorded_date': ['resource', 'recordedDate']
            }
        },
        'DiagnosticReport': {
            'table_meta': {
                'table_name': 'diagnostic_reports',
            },
            'json_schema': {
                'id': ['fullUrl'],
                'request_method': ['request', 'method'],
                'resource_type': ['resource', 'resourceType'],
                'resource_id': ['resource', 'id'],
                'status': ['resource', 'status'],
                'category': ['resource', 'category'],
                'code_coding': ['resource', 'code', 'coding'],
                'code': ['resource', 'code', 'text'],
                'subject_reference': ['resource', 'subject', 'reference'],
                'encounter_reference': ['resource', 'encounter', 'reference'],
                'effective_date_time': ['resource', 'effectiveDateTime'],
                'issued_date_time': ['resource', 'issued'],
                'performer': ['resource', 'performer'],
                'result': ['resource', 'result'],
                'presented_form': ['resource', 'presentedForm']
            }
        },
        'DocumentReference': {
            'table_meta': {
                'table_name': 'document_references',
            },
            'json_schema': {
                'id': ['fullUrl'],
                'request_method': ['request', 'method'],
                'resource_type': ['resource', 'resourceType'],
                'resource_id': ['resource', 'id'],
                'status': ['resource', 'status'],
                'type_coding': ['resource', 'type', 'coding'],
                'category': ['resource', 'category'],
                'subject_reference': ['resource', 'subject', 'reference'],
                'date': ['resource', 'date'],
                'author': ['resource', 'author'],
                'custodian_reference': ['resource', 'custodian', 'reference'],
                'custodian': ['resource', 'custodian', 'display'],
                'content': ['resource', 'content'],
                'context_encounter': ['resource', 'context', 'encounter'],
                'context_period_start': ['resource', 'context', 'period', 'start'],
                'context_period_end': ['resource', 'context', 'period', 'end']
            }
        },
        'Encounter': {
            'table_meta': {
                'table_name': 'encounters',
            },
            'json_schema': {
                'id': ['fullUrl'],
                'request_method': ['request', 'method'],
                'resource_type': ['resource', 'resourceType'],
                'resource_id': ['resource', 'id'],
                'status': ['resource', 'status'],
                'class_code': ['resource', 'class', 'code'],
                'type': ['resource', 'type'],
                'subject_reference': ['resource', 'subject', 'reference'],
                'subject': ['resource', 'subject', 'display'],
                'participant': ['resource', 'participant'],
                'period_start': ['resource', 'period', 'start'],
                'period_end': ['resource', 'period', 'end'],
                'reason_code': ['resource', 'reasonCode'],
                'hospitalization_discharge_disposition_coding': ['resource', 'hospitalization', 'dischargeDisposition', 'coding'],
                'hospitalization_discharge_disposition': ['resource', 'hospitalization', 'dischargeDisposition', 'text'],
                'location': ['resource', 'location'],
                'service_provider_reference': ['resource', 'serviceProvider', 'reference'],
                'service_provider': ['resource', 'serviceProvider', 'display']
            }
        },
        'ExplanationOfBenefit': {
            'table_meta': {
                'table_name': 'explanation_of_benefits',
            },
            'json_schema': {
                'id': ['fullUrl'],
                'request_method': ['request', 'method'],
                'resource_type': ['resource', 'resourceType'],
                'resource_id': ['resource', 'id'],
                'contained': ['resource', 'contained'],
                'identifier': ['resource', 'identifier'],
                'status': ['resource', 'status'],
                'type_coding': ['resource', 'type', 'coding'],
                'use': ['resource', 'use'],
                'patient_reference': ['resource', 'patient', 'reference'],
                'billable_period_start': ['resource', 'billablePeriod', 'start'],
                'billable_period_end': ['resource', 'billablePeriod', 'end'],
                'created': ['resource', 'created'],
                'insurer': ['resource', 'insurer', 'display'],
                'provider_reference': ['resource', 'provider', 'reference'],
                'referral_reference': ['resource', 'referral', 'reference'],
                'facility_reference': ['resource', 'facility', 'reference'],
                'facility': ['resource', 'facility', 'display'],
                'claim_reference': ['resource', 'claim', 'reference'],
                'outcome': ['resource', 'outcome'],
                'outcome': ['resource', 'careTeam'],
                'diagnosis': ['resource', 'diagnosis'],
                'insurance': ['resource', 'insurance'],
                'item': ['resource', 'item'],
                'total': ['resource', 'total'],
                'payment_amount_value': ['resource', 'payment' 'amount', 'value'],
                'payment_amount_currency': ['resource', 'payment', 'amount', 'currency']
            }
        },
        'MedicationRequest': {
            'table_meta': {
                'table_name': 'medication_requests',
            },
            'json_schema': {
                'id': ['fullUrl'],
                'request_method': ['request', 'method'],
                'resource_type': ['resource', 'resourceType'],
                'resource_id': ['resource', 'id'],
                'status': ['resource', 'status'],
                'intent': ['resource', 'intent'],
                'medication_reference': ['resource', 'medicationReference', 'reference'],
                'medication_coding': ['resource', 'medication', 'coding'],
                'medication': ['resource', 'medication', 'text'],
                'subject_reference': ['resource', 'subject', 'reference'],
                'encounter_reference': ['resource', 'encounter', 'reference'],
                'authored_on': ['resource', 'authoredOn'],
                'requester_reference': ['resource', 'requester', 'reference'],
                'requester': ['resource', 'requester', 'display'],
                'reason_reference': ['resource', 'reasonReference'],
                'dosage_instruction': ['resource', 'dosageInstruction'],
            }
        },
        'Patient': {
            'table_meta': {
                'table_name': 'patients',
            },
            'json_schema': {
                'id': ['fullUrl'],
                'request_method': ['request', 'method'],
                'resource_type': ['resource', 'resourceType'],
                'resource_id': ['resource', 'id'],
                'extension': ['resource', 'extension'],
                'identifier': ['resource', 'identifier'],
                'name': ['resource', 'name'],
                'contact': ['resource', 'telecom'],
                'gender': ['resource', 'gender'],
                'birth_date': ['resource', 'birthDate'],
                'deceased_date_time': ['resource', 'deceasedDateTime'],
                'address': ['resource', 'address'],
                'marital_status_coding': ['resource', 'maritalStatus', 'coding'],
                'marital_status': ['resource', 'maritalStatus', 'text'],
                'multiple_birth': ['resource', 'multipleBirthBoolean'],
                'multiple_birth_amount': ['resource', 'multipleBirthInteger'],
                'communication': ['resource', 'communication']
            }
        },
        'Procedure': {
            'table_meta': {
                'table_name': 'procedures',
            },
            'json_schema': {
                'id': ['fullUrl'],
                'request_method': ['request', 'method'],
                'resource_type': ['resource', 'resourceType'],
                'resource_id': ['resource', 'id'],
                'status': ['resource', 'status'],
                'code_coding': ['resource', 'code', 'coding'],
                'code': ['resource', 'code', 'text'],
                'subject_reference': ['resource', 'subject', 'reference'],
                'encounter_reference': ['resource', 'encounter', 'reference'],
                'performed_period_start': ['resource', 'performedPeriod', 'start'],
                'performed_period_end': ['resource', 'performedPeriod', 'end'],
                'location_reference': ['resource', 'location', 'reference'],
                'location': ['resource', 'location', 'display'],
                'reason_reference': ['resource', 'reasonReference'],
            }
        }
    }


"""
Utility file to make database connections.
"""


import psycopg2
import psycopg2.extensions


def create_connection(server_config: dict) -> psycopg2.extensions.connection:
    """
    Creates a connection to the database using the credentials provided by the config.ini file.
    If database_connection is False, connects to default database.

    :param server_config:
        The database server configuration dictionary which includes host, port, user, password, and database information.
    :param database_connection:
        True of False, whether or not to connect to the database in the config.ini file (otherwise connect to the default database).

    :returns:
        postgreSQL database server connection object
    """
    try:
        if server_config['DATABASE']:
            conn = psycopg2.connect(
                host=server_config['HOST'],
                port=server_config['PORT'],
                user=server_config['USER'],
                password=server_config['PASSWORD'],
                database=server_config['DATABASE']
            )
        else:
            conn = psycopg2.connect(
                host=server_config['HOST'],
                port=server_config['PORT'],
                user=server_config['USER'],
                password=server_config['PASSWORD']
            )
        print("Connection to server established.")
        return conn
    except psycopg2.OperationalError as e:
        raise Exception("Error connecting to postgreSQL server due to error: {e}")


"""
The main ETL pipeline file.
Extracts the JSON data, one resource at a time, from the JSON files in the specified file path.
Transforms the JSON data by converting each dictionary-like structure to a row of values.
Loads the rows generated by the transformations to the database tables in batches.
"""


import os
import sys
import argparse
import time

from src.database.database_functions.database_connection import create_connection
from src.etl.pipeline_functions import extract, transform, load
from src.etl.resources.schemas import SCHEMAS


def get_duplicate_entries_count(logs_dir):
    dup_str = 'duplicate key value'

    log_files = [os.path.join(logs_dir, file_name) for file_name in os.listdir(logs_dir)]
    log_file = sorted(log_files, key=os.path.getmtime, reverse=True)[0]

    count = 0
    with open(log_file, 'r') as f:
        for line in f:
            count += line.count(dup_str)
    
    return count


def pipeline(json_file_paths: list, server_config: dict):
    """
    Pipeline function creates a dictionary to hold the lists of all the rows for each resource type to load into the database.
    Calls the extract_resources() function to extract each resource one by one from all the files, loops over them, and transforms them to rows.
    Each returned row by transform_resource() is appended to the dictionary for that resrouce type.
    If any list reaches 1000 items, the 1000 rows are inserted to the table for that resource as a batch.
    Once the iterations are completed, any remaining rows in the dictionary are inserted.

    :param json_file_paths:
        List of strings of valid file paths of the JSON files to process.
    :param server_config:
        The database server configuration dictionary which includes host, port, user, password, and database information.
    """
    # Logging
    logs_dir = 'logs'
    if not os.path.exists(logs_dir):
        try:
            os.makedirs(logs_dir)
        except Exception as e:
            raise Exception(f"Could not create directory for logs due to error: {e}")
    num_logs = len(os.listdir(logs_dir))

    # Initialise dictionary to hold the rows of data.
    data_rows = {
        'CarePlan': [],
        'Claim': [],
        'Condition': [],
        'DiagnosticReport': [],
        'DocumentReference': [],
        'Encounter': [],
        'ExplanationOfBenefit': [],
        'MedicationRequest': [],
        'Patient': [],
        'Procedure': []
    }
    
    # Connect to database.
    conn = create_connection(server_config=server_config)
    conn.autocommit = False

    print(f"Loading {len(json_file_paths)} files to database.")
    start_time = time.time()
    
    # Resources count
    i = 0

    # Iterate over resources from the JSON data extracted from the files.
    for resource in extract.extract_resources(json_file_paths=json_file_paths):


        # Only transform resource types in dictionary.
        if resource['resource']['resourceType'] in data_rows:
            transformed_resource = transform.transform_resource(resource)
            
            resource_type = resource['resource']['resourceType']

            # Append the transformed row to the dictionary.
            data_rows[resource_type].append(transformed_resource)
        
            # If row amounts reach 1000, insert to database as batch.
            if len(data_rows[resource_type]) == 1000:
                table_name = SCHEMAS[resource_type]['table_meta']['table_name']
                num_cols = len(SCHEMAS[resource_type]['json_schema'])

                load.upload_resources(conn=conn, table_name=table_name, num_cols=num_cols, rows=data_rows[resource_type])

                # Empty list of rows for next batch.
                data_rows[resource_type] = []

            i += 1

    # Loop through the data if any left and insert the data as batches to table
    for resource_type, rows in data_rows.items():
        if rows:
            table_name = SCHEMAS[resource_type]['table_meta']['table_name']
            num_cols = len(SCHEMAS[resource_type]['json_schema'])

            load.upload_resources(conn=conn, table_name=table_name, num_cols=num_cols, rows=rows)
    
    end_time = time.time()

    conn.close()

    duplicates = get_duplicate_entries_count(logs_dir=logs_dir)

    if len(os.listdir(logs_dir)) > num_logs:
        i -= duplicates
        print("WARNING: There were some errors when inserting to database. Check log file.")
    
    print(f"{i} resources extracted, processed, and loaded to database in {end_time - start_time} seconds.")



def get_args() -> argparse.ArgumentParser:
    """
    Argument parser function for this file. Sets arguments and parses them for this file.

    :returns:
        Parsed arguments.
    """
    # Argument parser
    parser = argparse.ArgumentParser()

    # Add arguments
    parser.add_argument(
        '--host', type=str, required=True,
        help='The host name of the postgreSQL database server to connect to.'
    )
    parser.add_argument(
        '--port', type=str, required=True,
        help='The port of the postgreSQL database server to connect to.'
    )
    parser.add_argument(
        '--user', type=str, required=True,
        help='The username of the postgreSQL database server to connect to.'
    )
    parser.add_argument(
        '--password', type=str, required=True,
        help='The password of the postgreSQL database server to connect to.'
    )
    parser.add_argument(
        '--database', type=str, required=False, default=None,
        help='The database of the postgreSQL database server to connect to. If not specified, user will require access to default database.'
    )
    parser.add_argument(
        '--path', type=str, required=False, default=None,
        help='The path to the data directory.'
    )

    # Parse args
    args = parser.parse_args()

    return args


def main():
    """
    Running this file will insert all the processed data extracted and processed from each file to the database.
    """
    # Specify the directory holding the data.
    # Get args
    args = get_args()

    data_dir = args.path

    if not data_dir:
        if os.environ.get('DOCKER_CONTAINER'):
            data_dir = 'data/'
        else:
            print("""usage: fhir-load.py [-h] --path PATH --host HOST --port PORT --user USER
                        --password PASSWORD [--database DATABASE]
                \nthe following arguments are required: --path, --host, --port, --user, --password
            """)
            sys.exit(1)

    if not os.path.isdir(data_dir):
        print(f"{data_dir} is not a folder.")
        sys.exit(1)
    
    for file_path in os.listdir(data_dir):
        name, extension = os.path.splitext(file_path)
        if extension != '.json':
            print(f"{file_path} is not a JSON file.")
            sys.exit(1)
    
    json_file_paths = [os.path.join(data_dir, file_name) for file_name in os.listdir(data_dir)]
    
    server_config = {
        'HOST': args.host,
        'PORT': args.port,
        'USER': args.user,
        'PASSWORD': args.password,
        'DATABASE': args.database
    }
    print("Server credentials loaded.")
    
    # Run pipeline.
    pipeline(json_file_paths=json_file_paths, server_config=server_config)

if __name__ == '__main__':
    main()


"""
Transform function transform_resource() returns a newly created row which contains the values to insert to the table that the resource references.
It creates the row by extracting the values from JSON structured data, some of which is nested, by calling on the get_value_for_col() function.
"""


import json

from src.etl.resources.schemas import SCHEMAS


def get_value_for_col(resource: dict, path_to_value: list):
    """
    Iterative function to retrieve the required value by iterating into the nested layers of the resource dictionary.

    :param resource:
        JSON data containing values nested in that need to be extracted to insert to the table of the resource type.
    :param path_to_value:
        List containing the key path to the nested value within the JSON data.
    """
    # Initialise value to the full JSON data.
    value = resource

    # Iterate into the JSON data to retrieve the nested value required.
    for nested_key in path_to_value:
        if isinstance(value, dict):
            value = value.get(nested_key, 'null')

    # Reformatting necessary to match data types for SQL tables.        
    if isinstance(value, list):
        value = json.dumps(value, ensure_ascii=False)
    if isinstance(value, str):
        value = value.replace("'", "''")

    return value


def transform_resource(resource: dict) -> list:
    """
    Uses a stored 'schema' (JSON structure) for each resource type to extract the values needed for the columns in that resource table.
    Appends these values to a list to represent a row which will be inserted to the table.

    :param resource:
        JSON data containing values nested in that need to be extracted to insert to the table of the resource type.

    :returns:
        The row of extracted values as a list.
    """
    # Map for columns to the values nested in the JSON data.
    schema = SCHEMAS[resource['resource']['resourceType']]['json_schema']

    new_row = []

    # Iterate over each 'column' in the schema to get the value for that column.
    for col, path_to_value in schema.items():
        value = get_value_for_col(resource=resource, path_to_value=path_to_value)
        # Add the value to the row.
        new_row.append(value)

    return new_row


"""
create_database() function is used to create the fhir_database and its associated tables. The table names can be found in the tables.py file.
The function takes a configuration dictionary containing postgreSQL database server connection details.
It creates the database and tables by executing the SQL command in the database.py file and the tables.py file.
"""


import argparse

import psycopg2
import psycopg2.extensions
import psycopg2.errors

from src.database.database_functions.database_connection import create_connection
from src.database.database_sql import database, tables


def execute_sql_commands(conn: psycopg2.extensions.connection, sql: list):
    """
    Executes SQL commands specified in a file, and commits the execution.

    :param conn: psycopg2 connection to a postgreSQL sever database.
    :param sql: contents of the SQL file to be executed.
    """
    with conn.cursor() as cursor:
        for sql_statement in sql:
            try:
                cursor.execute(sql_statement)
            except (psycopg2.errors.DuplicateDatabase, psycopg2.errors.DuplicateTable) as e:
                conn.rollback()
                print(e)
            except psycopg2.errors.SyntaxError as e:
                conn.rollback()
                cursor.close()
                conn.close()
                raise Exception(f"Could not execute SQL statement\n{sql_statement}\ndue to error: {e}")
            sql_statement = ''
        conn.commit()


def create_database(server_config: dict):
    """
    Creates the database and associated tables in the default schema. The database is called fhir_database.
    The table names can be found in the tables.py file.

    :param server_config:
        The database server configuration dictionary which includes host, port, user, password, and database information.
    """
    # Creating database using database.py file.
    print("Creating fhir_database...")

    conn = create_connection(server_config=server_config)
    conn.autocommit = True

    execute_sql_commands(conn=conn, sql=database.database_sql)

    conn.close()
    print("Connection closed.")

    # Creating tables using tables.py file.
    print("Creating fhir_database tables...")
        
    server_config['DATABASE'] = 'fhir_database'
    conn = create_connection(server_config=server_config)
    
    execute_sql_commands(conn=conn, sql=tables.tables_sql)

    conn.close()
    print("Connection closed.")
    
    print("Database and tables successfully created.")


def get_args() -> argparse.ArgumentParser:
    """
    Argument parser function for this file. Sets arguments and parses them for this file.

    :returns:
        Parsed arguments.
    """
    # Argument parser
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--host', type=str, required=True,
        help='The host name of the postgreSQL database server to connect to.'
    )
    parser.add_argument(
        '--port', type=str, required=True,
        help='The port of the postgreSQL database server to connect to.'
    )
    parser.add_argument(
        '--user', type=str, required=True,
        help='The username of the postgreSQL database server to connect to.'
    )
    parser.add_argument(
        '--password', type=str, required=True,
        help='The password of the postgreSQL database server to connect to.'
    )
    parser.add_argument(
        '--database', type=str, required=False, default=None,
        help='The database of the postgreSQL database server to connect to. If not specified, user will require access to default database.'
    )

    # Parse args
    args = parser.parse_args()

    return args


def main():
    """
    Running this file will create the database and associated tables in the postgreSQL server whose details are provided in config.ini.
    """
    # Get args
    args = get_args()

    # Read postgreSQL configuration from config.ini file.
    server_config = {
        'HOST': args.host,
        'PORT': args.port,
        'USER': args.user,
        'PASSWORD': args.password,
        'DATABASE': args.database
    }
    print("Server credentials loaded.")

    # Create database and tables.
    create_database(server_config=server_config)
    

if __name__ == '__main__':
    main()



"""
Load function upload_resource() batch inserts rows into a table.
The function receives the number of columns to create the value_placeholders in the SQL query string.
"""


import os
import logging
import datetime

import psycopg2.errors
import psycopg2.extensions
from psycopg2.extras import execute_values

errors = False


def upload_resources(conn: psycopg2.extensions.connection, table_name: str, num_cols: int, rows: list):
    """
    Inserts the batch of rows to the table referenced using a dynamically formatted SQL string.
    
    :param conn:
        Database connection object.
    :param table_name:
        The table name to insert the records to.
    :param num_cols:
        The number of columns of the table.
    :param rows:
        The list of rows to insert into the table.
    """

    cursor = conn.cursor()

    # Try batch insert initially
    try:
        query = f"INSERT INTO {table_name} VALUES %s"
        execute_values(cursor, query, rows, page_size=1000)
    # Catch unique key errors
    except psycopg2.errors.UniqueViolation:
        conn.rollback()
        # Create number of value_placeholders from the number of columns.
        value_placeholders = ','.join(["'{}'"] *num_cols)

        # Iterate over the rows to insert.
        for row in rows:
            query = f"INSERT INTO {table_name} VALUES ({value_placeholders})".format(*row)
            try:
                cursor.execute(query)
            except psycopg2.errors.UniqueViolation as e:
                conn.rollback()
                global errors
                if not errors:
                    errors = True
                    log_file = os.path.join('logs', datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S.log"))
                    logging.basicConfig(filename=log_file, level=logging.ERROR)
                logging.error(f"Could not insert row to {table_name} due to error:\n{e}")
            except Exception as e:
                conn.rollback()
                cursor.close()
                conn.close()
                raise Exception(f"Could not execute query:\n{query}\nError:\n{e}")

    # Commit query executions.        
    conn.commit()
    
    cursor.close()


"""
Extract function extract_resources() is used to return the resources of each file.
"""


import json


def extract_resources(json_file_paths: list):
    """
    Iterates over the files in the file path and parses JSON data from each file.
    Iterates over the json data and startes generating the 'resources' in the 'entry' json object'.

    :param json_file_paths:
        List of strings of valid file paths of the JSON files to process.
    """
    for json_file_path in json_file_paths:
        try:
            with open(json_file_path, 'r', encoding='utf-8') as f:
                json_data = json.load(f)
                try:
                    for resource in json_data['entry']:
                        yield resource
                except (KeyError, TypeError) as e:
                    print(f"Could not extract resources: {e}")
        except (FileNotFoundError) as e:
            print(f"There was an error when opening file {json_file_path}: {e}")
        except json.JSONDecodeError as e:
            print(f"There was an error when decoding the file {json_file_path}: {e}")